# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
import logging
from subprocess import Popen, TimeoutExpired
import sys
from threading import Event, Thread
import time

from ..prim import DazlError, TimeDeltaLike, to_timedelta


def kill_process_tree(process: "Popen"):
    """
    Kill a process and its children.
    """
    #
    # noinspection PyBroadException
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import psutil

        children = psutil.Process(pid=process.pid).children()
    except:  # noqa
        children = []

    process.terminate()
    try:
        # give the process a few seconds to die
        process.communicate(timeout=5)

    except TimeoutExpired:
        logging.error("Had trouble killing a sandbox process normally; it will be forcibly killed.")
        # if there are child processes, start working around the parent process and try to
        # terminate them directly
        #
        # noinspection PyBroadException
        for child in children:
            try:
                child.kill()
            except:  # noqa
                pass

        process.kill()
        process.communicate()


def wait_for_process_port(process: "Popen", port: int, timeout: "TimeDeltaLike") -> None:
    from .io import is_port_alive

    alive = False
    max_time_wait = datetime.utcnow() + to_timedelta(timeout)

    logging.debug("Waiting for port %s to be alive on pid %s...", port, process.pid)
    while (
        (max_time_wait is None or (datetime.utcnow() < max_time_wait))
        and process.poll() is None
        and not alive
    ):
        alive = is_port_alive(port)
        if not alive:
            time.sleep(0.1)

    if not alive:
        return_code = process.returncode
        if return_code is not None:
            raise ProcessDiedException(return_code, "The process exited with an error code")
        raise Exception("Timed out while waiting for a process to start")


class ProcessLogger:
    """
    Pipe stdout and stderr from a process to the logger.
    """

    def __init__(self, process: "Popen", logger: "logging.Logger"):
        self.process = process
        self.logger = logger
        self._evt = Event()

    def start(self):
        stdout = sys.stdout
        stderr = sys.stderr
        stdout_log_thread = Thread(target=self._stdout_monitor, args=[stdout])
        stdout_log_thread.start()
        stderr_log_thread = Thread(target=self._stderr_monitor, args=[stderr])
        stderr_log_thread.start()

    def stop(self):
        self._evt.set()

    def _stdout_monitor(self, stdout):
        sys.stdout = stdout
        # noinspection PyBroadException
        try:
            for line in self.process.stdout:
                if self._evt.is_set():
                    return

                self.logger.info(line.rstrip("\n"))
        except:  # noqa
            pass

    def _stderr_monitor(self, stderr):
        sys.stderr = stderr
        # noinspection PyBroadException
        try:
            for line in self.process.stderr:
                if self._evt.is_set():
                    return

                self.logger.info(line.rstrip("\n"))
        except:  # noqa
            pass


class ProcessDiedException(DazlError):
    pass



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as _List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    TypeVar as __TypeVar,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.duration_pb2 import Duration
from google.protobuf.message import Message as __Message
from google.protobuf.timestamp_pb2 import Timestamp

__all__ = [
    "GetTimeModelRequest",
    "GetTimeModelResponse",
    "SetTimeModelRequest",
    "SetTimeModelResponse",
    "TimeModel",
]

class GetTimeModelRequest(__Message):
    def __init__(self) -> None: ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTimeModelRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Void) -> __Void: ...
    def ClearField(self, field_name: __Void) -> __Void: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetTimeModelResponse(__Message):
    configuration_generation: int
    time_model: TimeModel
    def __init__(self, *, configuration_generation: __Optional[int] = ..., time_model: __Optional[TimeModel] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTimeModelResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["configuration_generation", "time_model"]) -> bool: ...
    def ClearField(self, field_name: __Literal["configuration_generation", "time_model"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class SetTimeModelRequest(__Message):
    submission_id: str
    maximum_record_time: Timestamp
    configuration_generation: int
    new_time_model: TimeModel
    def __init__(self, *, submission_id: __Optional[int] = ..., maximum_record_time: __Optional[Timestamp] = ..., configuration_generation: __Optional[int] = ..., new_time_model: __Optional[TimeModel] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: SetTimeModelRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["submission_id", "maximum_record_time", "configuration_generation", "new_time_model"]) -> bool: ...
    def ClearField(self, field_name: __Literal["submission_id", "maximum_record_time", "configuration_generation", "new_time_model"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class SetTimeModelResponse(__Message):
    configuration_generation: int
    def __init__(self, *, configuration_generation: __Optional[int] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: SetTimeModelResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["configuration_generation"]) -> bool: ...
    def ClearField(self, field_name: __Literal["configuration_generation"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class TimeModel(__Message):
    avg_transaction_latency: Duration
    min_skew: Duration
    max_skew: Duration
    def __init__(self, *, avg_transaction_latency: __Optional[Duration] = ..., min_skew: __Optional[Duration] = ..., max_skew: __Optional[Duration] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: TimeModel) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["avg_transaction_latency", "min_skew", "max_skew"]) -> bool: ...
    def ClearField(self, field_name: __Literal["avg_transaction_latency", "min_skew", "max_skew"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import AbstractSet, BinaryIO, Collection, Generator, Mapping, Optional, Union
from zipfile import ZipFile

from .._gen.com.daml.daml_lf_dev import daml_lf_pb2 as pb
from .daml_lf_1 import Archive, Package, PackageRef
from .parse import parse_archive

# Wherever the API expects a DAR, we can take a file path, `bytes`, or a byte buffer.
Dar = Union[bytes, str, Path, BinaryIO]

__all__ = ["Dar", "DarFile", "CachedDarFile", "get_dar_package_ids"]


class DarFile:
    """
    Provides access to the contents of a .dar file.

    This conforms to the :class:`PackageProvider` protocol.

    This class is _not_ thread-safe.
    """

    filename: Optional[str]

    def __init__(self, dar: "Dar"):
        """
        Initialize a new DarFile.

        :param dar:
            A path to a file (either expressed as str or pathlib.Path), a bytes blob, or a buffer.
        """
        if isinstance(dar, str):
            self.filename = dar
            self._buf = None
            self._zip = ZipFile(self.filename)

        elif isinstance(dar, PathLike):
            self.filename = dar.__fspath__()
            self._buf = None
            self._zip = ZipFile(self.filename)

        elif isinstance(dar, bytes):
            self.filename = None
            self._buf = BytesIO(dar)
            self._zip = ZipFile(self._buf)

        elif hasattr(dar, "read"):
            # file-like object (buffers, files, etc.)
            self.filename = None
            self._buf = None
            self._zip = ZipFile(dar)

        else:
            raise TypeError("DarFile only understands file paths or binary blobs")

    def __enter__(self) -> "DarFile":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._zip.close()
        if self._buf is not None:
            self._buf.close()

    def manifest(self) -> "Optional[Mapping[str, str]]":
        """
        Return the contents of the manifest of this DAR.
        """
        names = self._zip.namelist()
        if "META-INF/MANIFEST.MF" in names:
            manifest_bytes = self._zip.read("META-INF/MANIFEST.MF")
            manifest = {}
            for line in manifest_bytes.decode("utf-8").splitlines():
                name, _, value = line.partition(":")
                manifest[name] = value.strip()
            return manifest
        else:
            return None

    def sdk_version(self) -> "Optional[str]":
        """
        Return the SDK version used to compile this dar (if this information is available).
        """
        manifest = self.manifest()
        return manifest.get("Sdk-Version") if manifest is not None else None

    def archives(self) -> "Collection[Archive]":
        """
        Return :class:`Archive` instances from this :class:`DarFile`.

        :class:`CachedDarFile` is a better choice than `DarFile` if this method is frequently called
        on the same :class:`DarFile`, as package parsing is an expensive operation.
        """
        return [parse_archive(PackageRef(a.hash), a.payload) for a in self._pb_archives()]

    def package(self, package_id: "PackageRef") -> "Package":
        """
        Return the :class:`Package` corresponding to the specified :class:`PackageRef`.

        This function is somewhat inefficient as it must check every DALF for the specified package
        ID. The :method:`DarFile.archives` function should be used if it is known that all archives
        in a DAR are to be accessed, or use a :class:`CachedDarFile` instance if random access of
        packages by package ID are needed.
        """
        payload = self.package_bytes(package_id)
        return parse_archive(package_id, payload).package

    def package_bytes(self, package_id: "PackageRef") -> bytes:
        """
        Return bytes corresponding to the specified :class:`PackageRef`. If this DAR were to be
        uploaded to a ledger, these are the bytes that would be returned for the specified
        :class:`PackageRef`.

        Note that this NOT the same as simply returning a ``.dalf`` file within a DAR, because a
        ``.dalf`` contains an envelope in the form of ``ArchivePayload`` messages, and it is the
        _contents_ of this message that are persisted by Ledger API implementations.
        """
        for a in self._pb_archives():
            if a.hash == package_id:
                return a.payload

        # We do not raise PackageNotFoundError here (even though it seems like it would be a more
        # apt error) because PackageNotFoundError implies the operation is retryable
        raise Exception(f"package not found in a DAR: {package_id!r}")

    def package_ids(self) -> "AbstractSet[PackageRef]":
        """
        Return the set of package IDs from this DAR.
        """
        return frozenset(PackageRef(a.hash) for a in self._pb_archives())

    async def get_package(self, package_id: "PackageRef") -> bytes:
        """
        Return bytes corresponding to the specified :class:`PackageRef`. If this DAR were to be
        uploaded to a ledger, these are the bytes that would be returned for the specified
        :class:`PackageRef`.

        This method is not actually async; it merely has an async signature to comply with the
        PackageLoader protocol.
        """
        return self.package_bytes(package_id)

    async def list_package_ids(self):
        """
        Return the set of package IDs from this DAR.

        This method is not actually async; it merely has an async signature to comply with the
        PackageLoader protocol.
        """
        return self.package_ids()

    def _dalf_names(self) -> "Generator[PackageRef, None, None]":
        """
        Return a generator over the names of DALF files in this DarFile.
        """
        return (PackageRef(name) for name in self._zip.namelist() if name.endswith(".dalf"))

    def _pb_archives(self) -> "Generator[pb.Archive, None, None]":
        """
        Return a generator over :class:`pb.Archive` instances. Crucially, the Protobuf messages
        contain DAML-LF as a ``bytes`` field that has not yet been parsed.
        """
        for name in self._dalf_names():
            contents = self._zip.read(name)

            a = pb.Archive()
            a.ParseFromString(contents)
            yield a


class CachedDarFile:
    """
    A caching variation of :class:`DarFile`.

    :class:`CachedDarFile` tries to do disk operations only once, and hold onto their results in
    memory. This will generally improve performance at the expense of increased memory usage.
    """

    def __init__(self, dar: "Dar"):
        self._dar = dar
        from threading import Lock

        self._lock = Lock()
        self._archives = None  # type: Optional[Collection[Archive]]

    def archives(self) -> "Collection[Archive]":
        if self._archives is None:
            with self._lock:
                if self._archives is None:
                    with DarFile(self._dar) as dar:
                        self._archives = dar.archives()
        return self._archives

    def package(self, package_id: "PackageRef") -> "Package":
        # TODO: This import needs to be local as long as the dazl.util.dar module still exists
        #  to avoid import cycles. Move this to the top of the file when dazl.util.dar is removed.
        from .errors import PackageNotFoundError

        for archive in self.archives():
            if archive.hash == package_id:
                return archive.package

        raise PackageNotFoundError(package_id)

    def package_ids(self) -> "AbstractSet[PackageRef]":
        return frozenset([archive.hash for archive in self.archives()])


def get_dar_package_ids(dar: "Dar") -> "AbstractSet[PackageRef]":
    """
    Return the package IDs for a DAR.
    """
    with DarFile(dar) as dar_file:
        return dar_file.package_ids()



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/completion.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.rpc import status_pb2 as google_dot_rpc_dot_status__pb2

from . import trace_context_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/ledger/api/v1/completion.proto",
    package="com.daml.ledger.api.v1",
    syntax="proto3",
    serialized_options=b"\n\026com.daml.ledger.api.v1B\024CompletionOuterClass\252\002\026Com.Daml.Ledger.Api.V1",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\'com/daml/ledger/api/v1/completion.proto\x12\x16\x63om.daml.ledger.api.v1\x1a*com/daml/ledger/api/v1/trace_context.proto\x1a\x17google/rpc/status.proto"\x9a\x01\n\nCompletion\x12\x12\n\ncommand_id\x18\x01 \x01(\t\x12"\n\x06status\x18\x02 \x01(\x0b\x32\x12.google.rpc.Status\x12\x16\n\x0etransaction_id\x18\x03 \x01(\t\x12<\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32$.com.daml.ledger.api.v1.TraceContextBG\n\x16\x63om.daml.ledger.api.v1B\x14\x43ompletionOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3',
    dependencies=[
        com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,
        google_dot_rpc_dot_status__pb2.DESCRIPTOR,
    ],
)


_COMPLETION = _descriptor.Descriptor(
    name="Completion",
    full_name="com.daml.ledger.api.v1.Completion",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="command_id",
            full_name="com.daml.ledger.api.v1.Completion.command_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status",
            full_name="com.daml.ledger.api.v1.Completion.status",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="transaction_id",
            full_name="com.daml.ledger.api.v1.Completion.transaction_id",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="trace_context",
            full_name="com.daml.ledger.api.v1.Completion.trace_context",
            index=3,
            number=1000,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=137,
    serialized_end=291,
)

_COMPLETION.fields_by_name["status"].message_type = google_dot_rpc_dot_status__pb2._STATUS
_COMPLETION.fields_by_name[
    "trace_context"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
DESCRIPTOR.message_types_by_name["Completion"] = _COMPLETION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Completion = _reflection.GeneratedProtocolMessageType(
    "Completion",
    (_message.Message,),
    {
        "DESCRIPTOR": _COMPLETION,
        "__module__": "com.daml.ledger.api.v1.completion_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Completion)
    },
)
_sym_db.RegisterMessage(Completion)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/ledger_configuration_service.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2

from . import trace_context_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/ledger/api/v1/ledger_configuration_service.proto",
    package="com.daml.ledger.api.v1",
    syntax="proto3",
    serialized_options=b"\n\026com.daml.ledger.api.v1B$LedgerConfigurationServiceOuterClass\252\002\026Com.Daml.Ledger.Api.V1",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n9com/daml/ledger/api/v1/ledger_configuration_service.proto\x12\x16\x63om.daml.ledger.api.v1\x1a*com/daml/ledger/api/v1/trace_context.proto\x1a\x1egoogle/protobuf/duration.proto"p\n\x1dGetLedgerConfigurationRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12<\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32$.com.daml.ledger.api.v1.TraceContext"k\n\x1eGetLedgerConfigurationResponse\x12I\n\x14ledger_configuration\x18\x01 \x01(\x0b\x32+.com.daml.ledger.api.v1.LedgerConfiguration"\\\n\x13LedgerConfiguration\x12\x39\n\x16max_deduplication_time\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03\x32\xa8\x01\n\x1aLedgerConfigurationService\x12\x89\x01\n\x16GetLedgerConfiguration\x12\x35.com.daml.ledger.api.v1.GetLedgerConfigurationRequest\x1a\x36.com.daml.ledger.api.v1.GetLedgerConfigurationResponse0\x01\x42W\n\x16\x63om.daml.ledger.api.v1B$LedgerConfigurationServiceOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3',
    dependencies=[
        com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,
    ],
)


_GETLEDGERCONFIGURATIONREQUEST = _descriptor.Descriptor(
    name="GetLedgerConfigurationRequest",
    full_name="com.daml.ledger.api.v1.GetLedgerConfigurationRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="ledger_id",
            full_name="com.daml.ledger.api.v1.GetLedgerConfigurationRequest.ledger_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="trace_context",
            full_name="com.daml.ledger.api.v1.GetLedgerConfigurationRequest.trace_context",
            index=1,
            number=1000,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=161,
    serialized_end=273,
)


_GETLEDGERCONFIGURATIONRESPONSE = _descriptor.Descriptor(
    name="GetLedgerConfigurationResponse",
    full_name="com.daml.ledger.api.v1.GetLedgerConfigurationResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="ledger_configuration",
            full_name="com.daml.ledger.api.v1.GetLedgerConfigurationResponse.ledger_configuration",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=275,
    serialized_end=382,
)


_LEDGERCONFIGURATION = _descriptor.Descriptor(
    name="LedgerConfiguration",
    full_name="com.daml.ledger.api.v1.LedgerConfiguration",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="max_deduplication_time",
            full_name="com.daml.ledger.api.v1.LedgerConfiguration.max_deduplication_time",
            index=0,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=384,
    serialized_end=476,
)

_GETLEDGERCONFIGURATIONREQUEST.fields_by_name[
    "trace_context"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETLEDGERCONFIGURATIONRESPONSE.fields_by_name[
    "ledger_configuration"
].message_type = _LEDGERCONFIGURATION
_LEDGERCONFIGURATION.fields_by_name[
    "max_deduplication_time"
].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
DESCRIPTOR.message_types_by_name["GetLedgerConfigurationRequest"] = _GETLEDGERCONFIGURATIONREQUEST
DESCRIPTOR.message_types_by_name["GetLedgerConfigurationResponse"] = _GETLEDGERCONFIGURATIONRESPONSE
DESCRIPTOR.message_types_by_name["LedgerConfiguration"] = _LEDGERCONFIGURATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetLedgerConfigurationRequest = _reflection.GeneratedProtocolMessageType(
    "GetLedgerConfigurationRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETLEDGERCONFIGURATIONREQUEST,
        "__module__": "com.daml.ledger.api.v1.ledger_configuration_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetLedgerConfigurationRequest)
    },
)
_sym_db.RegisterMessage(GetLedgerConfigurationRequest)

GetLedgerConfigurationResponse = _reflection.GeneratedProtocolMessageType(
    "GetLedgerConfigurationResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETLEDGERCONFIGURATIONRESPONSE,
        "__module__": "com.daml.ledger.api.v1.ledger_configuration_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetLedgerConfigurationResponse)
    },
)
_sym_db.RegisterMessage(GetLedgerConfigurationResponse)

LedgerConfiguration = _reflection.GeneratedProtocolMessageType(
    "LedgerConfiguration",
    (_message.Message,),
    {
        "DESCRIPTOR": _LEDGERCONFIGURATION,
        "__module__": "com.daml.ledger.api.v1.ledger_configuration_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.LedgerConfiguration)
    },
)
_sym_db.RegisterMessage(LedgerConfiguration)


DESCRIPTOR._options = None

_LEDGERCONFIGURATIONSERVICE = _descriptor.ServiceDescriptor(
    name="LedgerConfigurationService",
    full_name="com.daml.ledger.api.v1.LedgerConfigurationService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=479,
    serialized_end=647,
    methods=[
        _descriptor.MethodDescriptor(
            name="GetLedgerConfiguration",
            full_name="com.daml.ledger.api.v1.LedgerConfigurationService.GetLedgerConfiguration",
            index=0,
            containing_service=None,
            input_type=_GETLEDGERCONFIGURATIONREQUEST,
            output_type=_GETLEDGERCONFIGURATIONRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_LEDGERCONFIGURATIONSERVICE)

DESCRIPTOR.services_by_name["LedgerConfigurationService"] = _LEDGERCONFIGURATIONSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# NOTE TO IMPLEMENTORS
#
# This file contains only public API definitions, overloads, and documentation. (This file should
# be treated more like a C header file than anything else.) The bulk of the implementation is kept
# in _party_client.py.
#
# This file is repetitive and tedious, but written this way primarily so that static typing tools
# do the right thing. Python's ``typing`` library (and mypy) aren't quite expressive enough to allow
# for a more concise representation of the various flavors of the API. The unit test
# ``test_api_consistency.py`` verifies that these implementations are generally in sync with each
# other the way that the documentation says they are.
"""
This module contains the public API for interacting with the ledger from the perspective of a
specific party.
"""

from asyncio import Future, ensure_future, get_event_loop
from contextlib import contextmanager
from functools import partial, wraps
from logging import INFO
import os
from pathlib import Path
from typing import Any, Awaitable, BinaryIO, Collection, List, Optional, Tuple, Union
from uuid import uuid4
import warnings

from .. import LOG
from ..damlast import get_dar_package_ids
from ..damlast.daml_lf_1 import PackageRef
from ..damlast.pkgfile import Dar
from ..damlast.protocols import SymbolLookup
from ..metrics import MetricEvents
from ..prim import ContractData, ContractId, Party, TimeDeltaLike, to_party
from ..protocols.events import (
    ContractArchiveEvent,
    ContractCreateEvent,
    ContractExercisedEvent,
    InitEvent,
    PackagesAddedEvent,
    ReadyEvent,
    TransactionEndEvent,
    TransactionStartEvent,
)
from ..query import ContractMatch, is_match
from ..scheduler import RunLevel, validate_install_signal_handlers
from ..util.asyncio_util import await_then
from ..util.io import get_bytes
from ..util.tools import as_list
from ._base_model import CREATE_IF_MISSING, IfMissingPartyBehavior
from ._events import (
    AEventHandler,
    AEventHandlerDecorator,
    EventHandler,
    EventHandlerDecorator,
    fluentize,
)
from ._network_client_impl import _NetworkImpl
from ._party_client_impl import _PartyClientImpl
from .bots import Bot, BotCollection
from .commands import EventHandlerResponse
from .config import AnonymousNetworkConfig, NetworkConfig, PartyConfig
from .events import EventKey
from .ledger import LedgerMetadata
from .state import ContractContextualData, ContractContextualDataCollection, ContractsState

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from ..model.types import TemplateNameLike

__all__ = [
    "DEFAULT_TIMEOUT_SECONDS",
    "simple_client",
    "async_network",
    "Network",
    "GlobalClient",
    "AIOGlobalClient",
    "SimpleGlobalClient",
    "PartyClient",
    "AIOPartyClient",
    "SimplePartyClient",
]
DEFAULT_TIMEOUT_SECONDS = 30


@contextmanager
def simple_client(
    url: Optional[str] = None,
    party: Union[None, str, Party] = None,
    log_level: Optional[int] = INFO,
):
    """
    Start up a single client connecting to a single specific party.

    :param url:
        The URL of the client to connect to. Defaults to the value of the ``DAML_LEDGER_URL``
        environment variable (if set).
    :param party:
        The party to connect as. Defaults to the value of the ``DAML_LEDGER_PARTY`` environment
        variable if it is set.
    :param log_level:
        If non-``None``, configure a default logger that logs output at the specified level. The
        default value is ``INFO``.
    :return:
        A :class:`SimplePartyClient` that can be used in a completely blocking, synchronous
        fashion.
    """
    if log_level is not None:
        from .. import setup_default_logger

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            setup_default_logger(log_level)

    if url is None:
        url = os.getenv("DAML_LEDGER_URL")
    if not url:
        raise ValueError(
            "url must be specified, or the DAML_LEDGER_URL environment variable " "must be set"
        )
    if not party:
        raise ValueError(
            "party must be specified, or the DAML_LEDGER_PARTY environment variable " "must be set"
        )

    LOG.info("Starting a simple_client with to %s with party %r...", url, party)

    network = Network()
    network.set_config(url=url)
    client = network.simple_party(party) if party else network.simple_new_party()

    network.start_in_background()

    yield client

    network.shutdown()
    network.join()


# This class is intended to be used as a function.
# noinspection PyPep8Naming
class async_network:
    """
    Create a :class:`Network` and ensure that it has the given set of DARs loaded.
    """

    def __init__(
        self,
        url: Optional[str] = None,
        dars: Optional[Union[Dar, Collection[Dar]]] = None,
    ):
        LOG.debug("async_network.__init__")
        self.network = Network()
        if url:
            self.network.set_config(url=url)
        self.dars = as_list(dars)  # type: List[Dar]

        LOG.debug("Analyzing package_id config...")
        self.package_ids = (
            {pkg_id for dar in self.dars for pkg_id in get_dar_package_ids(dar)}
            if self.dars
            else None
        )
        LOG.debug("Package id analysis done.")

        if self.package_ids:
            self.network.set_config(package_ids=self.package_ids)

    async def __aenter__(self):
        LOG.debug("async_network.__aenter__")

        for dar in self.dars:
            await self.network.aio_global().ensure_dar(dar)
        return self.network

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        LOG.debug("async_network.__aexit__")
        fut = self.network.shutdown()
        if fut is not None:
            await fut


class Network:
    """
    Manages network connection/scheduling logic on behalf of one or more :class:`PartyClient`
    instances.
    """

    def __init__(self, metrics: Optional[MetricEvents] = None):
        self._impl = _NetworkImpl(metrics)
        self._main_fut = None  # type: Optional[Future]

    def set_config(
        self,
        *config: Union[NetworkConfig, AnonymousNetworkConfig],
        url: Optional[str] = None,
        admin_url: Optional[str] = None,
        **kwargs
    ):
        self._impl.set_config(*config, url=url, admin_url=admin_url, **kwargs)

    @property
    def lookup(self) -> SymbolLookup:
        """
        Return a :class:`SymbolLookup` that provides type and package information for known
        packages.
        """
        return self._impl.lookup

    def resolved_config(self) -> NetworkConfig:
        """
        Calculate the configuration that will be used for this client when it is instantiated.
        """
        return self._impl.resolved_config()

    # <editor-fold desc="Global/Party client creation">

    def simple_global(self) -> "SimpleGlobalClient":
        """
        Return a :class:`GlobalClient` that exposes thread-safe, synchronous (blocking) methods for
        communicating with a ledger. Callbacks are dispatched to background threads.
        """
        return self._impl.global_impl(SimpleGlobalClient)

    def aio_global(self) -> "AIOGlobalClient":
        """
        Return a :class:`GlobalClient` that works on an asyncio event loop.

        Note that once this object can only be accessed from the asyncio event loop it is intended
        to be used on.
        """
        client = self._impl.global_impl(AIOGlobalClient)
        self._impl.freeze()
        return client

    def simple_party(self, party: Union[str, Party]) -> "SimplePartyClient":
        """
        Return a :class:`PartyClient` that exposes thread-safe, synchronous (blocking) methods for
        communicating with a ledger. Callbacks are dispatched to background threads.

        :param party: The party to get a client for.
        """
        return self._impl.party_impl_wrapper(to_party(party), SimplePartyClient)

    def simple_new_party(self) -> "SimplePartyClient":
        """
        Return a :class:`PartyClient` that exposes thread-safe, synchronous (blocking) methods for
        communicating with a ledger. Callbacks are dispatched to background threads.
        """
        return self.simple_party(str(uuid4()))

    def aio_party(self, party: Union[str, Party]) -> "AIOPartyClient":
        """
        Return a :class:`PartyClient` that works on an asyncio event loop.

        :param party: The party to get a client for.
        """
        return self._impl.party_impl_wrapper(Party(party), AIOPartyClient)

    def aio_new_party(self) -> "AIOPartyClient":
        """
        Return a :class:`PartyClient` for a random party that works on an asyncio event loop.
        This will never return the same object twice.
        """
        return self.aio_party(str(uuid4()))

    def party_bots(
        self, party: Union[str, Party], if_missing: IfMissingPartyBehavior = CREATE_IF_MISSING
    ) -> BotCollection:
        """
        Return the collection of bots associated with a party.

        :param party: The party to get bots for.
        :param if_missing:
            Specify the behavior to use in the case where no client has been yet requested for this
            party. The default behavior is CREATE_IF_MISSING.
        """
        warnings.warn(
            "bot introspection is deprecated and will be removed in dazl v8",
            DeprecationWarning,
            stacklevel=2,
        )
        party_impl = self._impl.party_impl(party, if_missing=if_missing)
        return party_impl.bots if party_impl is not None else None

    # </editor-fold>

    # <editor-fold desc="Daemon thread-based scheduling API">

    def start_in_background(
        self, daemon: bool = True, install_signal_handlers: Optional[bool] = None
    ) -> None:
        """
        Connect to the ledger in a background thread.

        The current thread does NOT block. Operations on instances of :class:`SimplePartyClient`
        are allowed, and operations on instances of :class:`AIOPartyClient` are allowed as long as
        they are made from the correct thread.
        """
        if validate_install_signal_handlers(install_signal_handlers):
            self._impl.invoker.install_signal_handlers()
        return self._impl.start(daemon)

    def shutdown(self) -> Optional[Awaitable[None]]:
        """
        Gracefully shut down all network connections and notify all clients that they are about to
        be terminated.

        The current thread does NOT block.

        :return:
            ``None`` unless ``start()`` was called, in which case the coroutine that
            corresponds to dazl's "main" is returned.
        """
        self._impl.shutdown()
        if self._main_fut is not None:
            warnings.warn(
                "shutdown() in conjunction with start() is deprecated; use aio_run() directly "
                "and await on the returned coroutine instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return self._main_fut
        else:
            return None

    def join(self, timeout: Optional[float] = None) -> None:
        """
        Block the current thread until the client is shut down.

        :param timeout:
            Number of seconds to wait before timing out the join, or ``None`` to wait indefinitely.
        """
        return self._impl.join(timeout=timeout)

    # </editor-fold>

    # <editor-fold desc="asyncio-based scheduling API">

    def start(self) -> None:
        """
        Start the coroutine that spawns callbacks for listeners on event streams.
        """
        warnings.warn(
            "start() is deprecated; use aio_run() directly", DeprecationWarning, stacklevel=2
        )
        self._main_fut = ensure_future(self.aio_run(keep_open=False))

    def run_until_complete(
        self, *coroutines: Awaitable[None], install_signal_handlers: Optional[bool] = None
    ) -> None:
        """
        Block the main thread and run the application in an event loop on the main thread. The loop
        terminates when the given (optional) coroutines terminate OR :meth:`shutdown` is called AND
        all active command submissions and event handlers' follow-ups have successfully returned.

        :param coroutines:
            Coroutines to run alongside event handlers and command submissions. When these
            coroutines are done running and the
        :param install_signal_handlers:
            ``True`` to install SIGINT and SIGQUIT event handlers (CTRL+C and CTRL+\\);
            ``False`` to skip installation. The default value is ``None``, which installs signal
            handlers only when called from the main thread (default). If signal handlers are
            requested to be installed and the thread is NOT the main thread, this method throws.
        """
        self._impl.invoker.level = RunLevel.RUN_UNTIL_IDLE
        self._impl.invoker.loop = get_event_loop()
        if validate_install_signal_handlers(install_signal_handlers):
            self._impl.invoker.install_signal_handlers()
        self._impl.invoker.loop.run_until_complete(self.aio_run(*coroutines))
        LOG.info("The internal run_until_complete event loop has now completed.")

    def run_forever(
        self, *coroutines: "Awaitable[None]", install_signal_handlers: "Optional[bool]" = None
    ) -> None:
        """
        Block the main thread and run the application in an event loop on the main thread. The loop
        terminates when :meth:`shutdown` is called AND all active command submissions and event
        handlers' follow-ups have successfully returned.
        """
        self._impl.invoker.loop = get_event_loop()
        if validate_install_signal_handlers(install_signal_handlers):
            self._impl.invoker.install_signal_handlers()
        self._impl.invoker.loop.run_until_complete(self.aio_run(*coroutines))
        LOG.info("The internal run_forever event loop has been shut down.")

    async def aio_run(self, *coroutines, keep_open: bool = True) -> None:
        """
        Coroutine where all network activity is scheduled from. This coroutine exits when
        :meth:`shutdown` is called, and can be used directly as an asyncio-native alternative to
        :meth:`start_in_background` and :meth:`join`.

        You would normally call this method directly only if you are trying to incorporate
        the client into an already-running event loop. Prefer :meth:`run_until_complete` or
        :meth:`run_forever` if you can block the current thread, or :meth:`start_in_background`
        with :meth:`join` if you wish to run the entire client on background threads.
        """
        if not keep_open:
            self._impl.invoker.level = RunLevel.RUN_UNTIL_IDLE
        await self._impl.aio_run(*coroutines)
        LOG.info("The aio_run coroutine has completed.")

    # </editor-fold>

    def parties(self) -> Collection[Party]:
        """
        Return a snapshot of the set of parties that exist right now.
        """
        return self._impl.parties()

    def bots(self) -> Collection[Bot]:
        """
        Return a collection of bots.

        Note that bot introspection will not be a part of the dazl v8 API.
        """
        warnings.warn(
            "bot introspection is deprecated and will be removed in dazl v8",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._impl.bots

    def __enter__(self):
        """
        Allows for use of a :class:`Network` as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()


class GlobalClient:
    """
    Public interface for either an async-based or a thread-safe version of an API for interacting
    with a Ledger API implementation that manages global ledger data, such as package store
    management and current time.
    """

    def __init__(self, impl: _NetworkImpl):
        self._impl = impl


class AIOGlobalClient(GlobalClient):
    async def ensure_dar(
        self,
        contents: Union[str, Path, bytes, BinaryIO],
        timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        Validate that the ledger has the packages specified by the given contents (as a byte array).
        Throw an exception if the specified DARs do not exist within the specified timeout.

        :param contents: The DAR or DALF to ensure.
        :param timeout: The maximum length of time to wait before giving up.
        """
        raw_bytes = get_bytes(contents)
        return await self._impl.upload_package(raw_bytes, timeout)

    async def ensure_packages(
        self,
        package_ids: Collection[PackageRef],
        timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        Validate that packages with the specified package IDs exist on the ledger. Throw an
        exception if the specified packages do not exist within the specified timeout.

        :param package_ids: The set of package IDs to check for.
        :param timeout: The maximum length of time to wait before giving up.
        """
        return await self._impl.ensure_package_ids(package_ids, timeout)

    async def metadata(self) -> LedgerMetadata:
        """
        Return the current set of known packages.
        """
        return await self._impl.aio_metadata()


class SimpleGlobalClient(GlobalClient):
    def ensure_dar(
        self,
        contents: Union[str, Path, bytes, BinaryIO],
        timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        Validate that the ledger has the packages specified by the given contents (as a byte array).
        Throw an exception if the specified DARs do not exist within the specified timeout.

        :param contents: The DAR or DALF to ensure.
        :param timeout: The maximum length of time to wait before giving up.
        """
        raw_bytes = get_bytes(contents)
        return self._impl.invoker.run_in_loop(lambda: self._impl.upload_package(raw_bytes, timeout))

    def ensure_packages(
        self,
        package_ids: Collection[PackageRef],
        timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        Validate that packages with the specified package IDs exist on the ledger. Throw an
        exception if the specified packages do not exist within the specified timeout.

        :param package_ids: The set of package IDs to check for.
        :param timeout: The maximum length of time to wait before giving up.
        """
        return self._impl.invoker.run_in_loop(
            lambda: self._impl.ensure_package_ids(package_ids, timeout)
        )

    def metadata(self, timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS) -> LedgerMetadata:
        """
        Return the current set of known packages.
        """
        return self._impl.simple_metadata(timeout)


class PartyClient:
    """
    Public interface for either an async-based or a thread-safe version of an API for interacting
    with a Ledger API implementation from the perspective of a single client.
    """

    def __init__(self, impl: _PartyClientImpl):
        self._impl = impl

    # <editor-fold desc="Ledger/client metadata">

    @property
    def party(self) -> Party:
        """
        Return the party serviced by this client.
        """
        return self._impl.party

    def resolved_config(self) -> PartyConfig:
        """
        Calculate the configuration that will be used for this client when it is instantiated.
        """
        return self._impl.resolved_config()

    # </editor-fold>


class AIOPartyClient(PartyClient):
    """
    Implementation of a :class:`PartyClient` that exposes an `async`/`await`-style API that runs on
    an event loop.
    """

    # <editor-fold desc="Event handler registration">

    def ledger_init(self) -> AEventHandlerDecorator[InitEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` has been
        instructed to begin, but before any network activity is started.
        """
        return fluentize(self.add_ledger_init)

    def add_ledger_init(self, handler: AEventHandler[InitEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` has been instructed to
        begin, but before any network activity is started.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        """
        for key in EventKey.init():
            self._impl.add_event_handler(key, handler, None, self)

    def ledger_ready(self) -> AEventHandlerDecorator[ReadyEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` has caught
        up to the head of the ledger, but before any :meth:`ledger_create` or :meth:`ledger_archive`
        callbacks are invoked.
        """
        return fluentize(self.add_ledger_ready)

    def add_ledger_ready(self, handler: AEventHandler[ReadyEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` has caught up to the head of
        the ledger, but before any :meth:`ledger_create` or :meth:`ledger_archive` callbacks are
        invoked.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        """
        for key in EventKey.ready():
            self._impl.add_event_handler(key, handler, None, self)

    def ledger_packages_added(
        self, initial: bool = False
    ) -> AEventHandlerDecorator[PackagesAddedEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` has
        detected new packages added to the ledger.

        :param initial:
            ``True`` to call the handler when the client is ready. This can be useful if you want
            to handle package additions identically whether they were already in the ledger when
            the client started up or only after a package has been added. The default value is
            ``False``, which means that this handler is only called on NEW packages that have been
            uploaded after this client has started.
        :return:
        """
        return fluentize(self.add_ledger_packages_added, initial=initial)

    def add_ledger_packages_added(
        self, handler: AEventHandler[PackagesAddedEvent], initial: bool = False
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` has detected new packages
        added to the ledger.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        :param initial:
            ``True`` to call the handler when the client is ready. This can be useful if you want
            to handle package additions identically whether they were already in the ledger when
            the client started up or only after a package has been added. The default value is
            ``False``, which means that this handler is only called on NEW packages that have been
            uploaded after this client has started.
        :return:
        """
        for key in EventKey.packages_added(initial=initial, changed=True):
            self._impl.add_event_handler(key, handler, None, self)

    def ledger_transaction_start(self) -> AEventHandlerDecorator[TransactionStartEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` receives a
        new transaction. Called before individual :meth:`ledger_create` and :meth:`ledger_archive`
        callbacks.
        """
        return fluentize(self.add_ledger_transaction_start)

    def add_ledger_transaction_start(self, handler: AEventHandler[TransactionStartEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` receives a new transaction.
        Called before individual :meth:`ledger_create` and :meth:`ledger_archive` callbacks.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        """
        for key in EventKey.transaction_start():
            self._impl.add_event_handler(key, handler, None, self)

    def ledger_transaction_end(self) -> AEventHandlerDecorator[TransactionEndEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` receives a
        new transaction. Called after individual :meth:`ledger_create` and :meth:`ledger_archive`
        callbacks.
        """
        return fluentize(self.add_ledger_transaction_end)

    def add_ledger_transaction_end(self, handler: AEventHandler[TransactionEndEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` receives a new transaction.
        Called after individual :meth:`ledger_create` and :meth:`ledger_archive` callbacks.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        """
        for key in EventKey.transaction_end():
            self._impl.add_event_handler(key, handler, None, self)

    def ledger_created(
        self, template: Any, match: Optional[ContractMatch] = None
    ) -> AEventHandlerDecorator[ContractCreateEvent]:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters a newly created
        template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """

        def _register_created(
            cb: AEventHandler[ContractCreateEvent],
        ) -> AEventHandler[ContractCreateEvent]:
            self.add_ledger_created(template, match=match, handler=cb)
            return cb

        return _register_created

    def add_ledger_created(
        self,
        template: Any,
        handler: AEventHandler[ContractCreateEvent],
        match: Optional[ContractMatch] = None,
    ) -> Bot:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters a newly created
        contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param handler:
            The callback to invoke whenever a matching template is created.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """
        filter_fn = partial(is_match, match) if match is not None else None

        bot = self._impl.bots.add_new(party_client=self, name=handler.__name__)
        bot.add_event_handler(EventKey.contract_created(True, template), handler, filter_fn)
        return bot

    def ledger_exercised(
        self, template: Any, choice: str
    ) -> AEventHandlerDecorator[ContractExercisedEvent]:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters an exercised
        choice event.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param choice:
            The name of the choice to listen for exercises on.
        """

        def _register_exercised(
            cb: AEventHandler[ContractExercisedEvent],
        ) -> AEventHandler[ContractExercisedEvent]:
            self.add_ledger_exercised(template, choice, handler=cb)
            return cb

        return _register_exercised

    def add_ledger_exercised(
        self, template: Any, choice: str, handler: AEventHandler[ContractExercisedEvent]
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters an exercised
        choice event.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param choice:
            The name of the choice to listen for exercises on.
        :param handler:
            The callback to invoke whenever a matching template is exercised.
        """
        for key in EventKey.contract_exercised(True, template, choice):
            self._impl.add_event_handler(key, handler, None, self)

    def ledger_archived(
        self, template: Any, match: Optional[ContractMatch] = None
    ) -> AEventHandlerDecorator[ContractArchiveEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` encounters
        a newly archived contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """

        def _register_archived(
            cb: AEventHandler[ContractArchiveEvent],
        ) -> AEventHandler[ContractArchiveEvent]:
            self.add_ledger_archived(template, match=match, handler=cb)
            return cb

        return _register_archived

    def add_ledger_archived(
        self,
        template: Any,
        handler: AEventHandler[ContractArchiveEvent],
        match: Optional[ContractMatch] = None,
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters a newly archived
        contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param handler:
            The callback to invoke whenever a matching template is created.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """
        filter_fn = partial(is_match, match) if match is not None else None

        for key in EventKey.contract_archived(True, template):
            self._impl.add_event_handler(key, handler, filter_fn, self)

    # </editor-fold>

    # <editor-fold desc="Command submission">

    def submit(
        self,
        commands: EventHandlerResponse,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> Awaitable[None]:
        """
        Submit commands to the ledger.

        :param commands:
            An object that can be converted to a command.
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        :return:
            A future that resolves when the command has made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        return self._impl.write_commands(
            commands, workflow_id=workflow_id, deduplication_time=deduplication_time
        )

    def submit_create(
        self,
        template_name: TemplateNameLike,
        arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> Awaitable[None]:
        """
        Submit a single create command. Equivalent to calling :meth:`submit` with a single
        ``create``.

        :param template_name:
            The name of the template.
        :param arguments:
            The arguments to the create (as a ``dict``).
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        :return:
            A future that resolves when the command has made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        from .. import create

        return self.submit(
            create(template_name, arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    def submit_exercise(
        self,
        cid: ContractId,
        choice_name: str,
        arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> Awaitable[None]:
        """
        Submit a single exercise choice. Equivalent to calling :meth:`submit` with a single
        ``exercise``.

        :param cid:
            The :class:`ContractId` on which a choice is being exercised.
        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            The arguments to the exercise (as a ``dict``). Can be omitted (``None``) for no-argument
            choices.
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        :return:
            A future that resolves when the command has made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        from .. import exercise

        return self.submit(
            exercise(cid, choice_name, arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    def submit_exercise_by_key(
        self,
        template_name: TemplateNameLike,
        contract_key: Any,
        choice_name: str,
        arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> Awaitable[None]:
        """
        Synchronously submit a single exercise choice. Equivalent to calling :meth:`submit` with a
        single ``exercise_by_key``.

        :param template_name:
            The name of the template on which to do an exercise-by-key.
        :param contract_key:
            The value that should uniquely identify a contract for the specified template.
        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            The arguments to the create (as a ``dict``). Can be omitted (``None``) for no-argument
            choices.
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        """
        from .. import exercise_by_key

        return self.submit(
            exercise_by_key(template_name, contract_key, choice_name, arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    def submit_create_and_exercise(
        self,
        template_name: TemplateNameLike,
        arguments: dict,
        choice_name: str,
        choice_arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> Awaitable[None]:
        """
        Synchronously submit a single create-and-exercise command. Equivalent to calling
        :meth:`submit` with a single ``create_and_exercise``.

        :param template_name:
            The name of the template on which to do an exercise-by-key.
        :param arguments:
            The arguments to the create (as a ``dict``).
        :param choice_name:
            The name of the choice to exercise.
        :param choice_arguments:
            The arguments to the exercise (as a ``dict``). Can be omitted (``None``) for no-argument
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        """
        from .. import create_and_exercise

        return self.submit(
            create_and_exercise(template_name, arguments, choice_name, choice_arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    # </editor-fold>

    # <editor-fold desc="Active contract set">

    def find_by_id(self, cid: Union[str, ContractId]) -> Optional[ContractContextualData]:
        return self._impl.find_by_id(cid)

    def find(
        self, template: Any, match: ContractMatch = None, include_archived: bool = False
    ) -> ContractContextualDataCollection:
        return self._impl.find(template, match, include_archived=include_archived)

    def find_active(self, template: Any, match: ContractMatch = None) -> ContractsState:
        """
        Immediately return data from the current active contract set.

        The contents of this ACS are guaranteed to be present (or removed) in the current
        transaction _before_ processing any corresponding ``on_created`` or ``on_archived``
        callbacks for this party. The ACS is populated _before_ processing any ``on_ready``
        callbacks.

        This method raises an error if ACS tracking has been disabled on this client.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :return:
            A ``dict`` whose keys are :class:`ContractId` and values are corresponding contract
            data that match the current query.
        """
        return self._impl.find_active(template, match)

    def find_historical(
        self, template: Any, match: ContractMatch = None
    ) -> ContractContextualDataCollection:
        """
        Immediately return data from the current active and historical contract set as
        a contextual data collection

        The contents of this set are guaranteed to be up-to-date in the current transaction _before_
        processing any corresponding ``on_created`` or ``on_archived`` callbacks for this party. The
        set is up-to-date _before_ processing any ``on_ready`` callbacks.

        This method raises an error if historical tracking has been disabled on this client.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :return:
            A ``ContractContextualDataCollection`` whose values correspond to the contract
            data for active and archived contracts matching the current query.
        """
        return self._impl.find_historical(template, match)

    def find_one(
        self, template: Any, match: "ContractMatch" = None, timeout: float = DEFAULT_TIMEOUT_SECONDS
    ) -> Awaitable[Tuple[ContractId, ContractData]]:
        """
        Return data from the current active contract set when at least some amount of rows exist in
        the active contract set.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :param timeout:
            Number of seconds in which to time out the search.
        :return:
            A ``Future`` that is resolved with a ``dict`` whose keys are :class:`ContractId` and
            values are corresponding contract data that match the current query.
        """
        return await_then(
            self.find_nonempty(template, match, min_count=1, timeout=timeout),
            lambda state: next(iter(state.items())),
        )

    def find_nonempty(
        self,
        template: Any,
        match: ContractMatch,
        min_count: int = 1,
        timeout: float = DEFAULT_TIMEOUT_SECONDS,
    ) -> Awaitable[ContractsState]:
        """
        Return data from the current active contract set when at least some amount of rows exist in
        the active contract set.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :param min_count:
            The minimum number of rows to return. The default value is 1.
        :param timeout:
            Number of seconds in which to time out the search.
        :return:
            A ``Future`` that is resolved with a ``dict`` whose keys are :class:`ContractId` and
            values are corresponding contract data that match the current query.
        """
        return self._impl.find_nonempty(template, match, min_count, timeout)

    # </editor-fold>

    # <editor-fold desc="Ledger/client metadata">

    def set_config(self, url: "Optional[str]", **kwargs):
        self._impl.set_config(url=url, **kwargs)

    async def ensure_dar(
        self,
        contents: Union[str, Path, bytes, BinaryIO],
        timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        Validate that the ledger has the packages specified by the given contents (as a byte array).
        Throw an exception if the specified DARs do not exist within the specified timeout.

        :param contents: The DAR or DALF to ensure.
        :param timeout: The maximum length of time to wait before giving up.
        """
        raw_bytes = get_bytes(contents)
        return await self._impl.parent.upload_package(raw_bytes, timeout)

    def ready(self) -> "Awaitable[None]":
        """
        Block until the ledger client has caught up to the current head and is ready to send
        commands.
        """
        return self._impl.ready()

    # </editor-fold>


class SimplePartyClient(PartyClient):
    """
    Implementation of a :class:`PartyClient` that exposes blocking calls, but can be used from any
    thread.

    Use this implementation if any of these apply:
      * you wish to interact with libraries that do not natively support asyncio
      * you are comfortable with the trade-off of having to block threads in order to write code
    """

    # <editor-fold desc="Event handler registration">

    def ledger_init(self) -> "EventHandlerDecorator[InitEvent]":
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` has been
        instructed to begin, but before any network activity is started.
        """
        return fluentize(self.add_ledger_init)

    def add_ledger_init(self, handler: EventHandler[InitEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` has been instructed to
        begin, but before any network activity is started.

        :param handler:
            The handler to register. May return anything that can be successfully coerced into a
            :class:`CommandPayload`.
        """

        @wraps(handler)
        def _background_ledger_init(event: InitEvent) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        for key in EventKey.init():
            self._impl.add_event_handler(key, _background_ledger_init, None, self)

    def ledger_ready(self) -> EventHandlerDecorator[ReadyEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` has caught
        up to the head of the ledger, but before any :meth:`ledger_create` or :meth:`ledger_archive`
        callbacks are invoked.
        """
        return fluentize(self.add_ledger_ready)

    def add_ledger_ready(self, handler: EventHandler[ReadyEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` has caught up to the head of
        the ledger, but before any :meth:`ledger_create` or :meth:`ledger_archive` callbacks are
        invoked.

        :param handler:
            The handler to register. May return anything that can be successfully coerced into a
            :class:`CommandPayload`.
        """

        @wraps(handler)
        def _background_ledger_ready(event: ReadyEvent) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        for key in EventKey.ready():
            self._impl.add_event_handler(key, _background_ledger_ready, None, self)

    def ledger_packages_added(
        self, initial: bool = False
    ) -> EventHandlerDecorator[PackagesAddedEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` has
        detected new packages added to the ledger.

        :param initial:
            ``True`` to call the handler when the client is ready. This can be useful if you want
            to handle package additions identically whether they were already in the ledger when
            the client started up or only after a package has been added. The default value is
            ``False``, which means that this handler is only called on NEW packages that have been
            uploaded after this client has started.
        :return:
        """
        return fluentize(self.add_ledger_packages_added, initial=initial)

    def add_ledger_packages_added(
        self, handler: EventHandler[PackagesAddedEvent], initial: bool = False
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` has detected new packages
        added to the ledger.

        :param handler:
            The handler to register. May return anything that can be successfully coerced into a
            :class:`CommandPayload`.
        :param initial:
            ``True`` to call the handler when the client is ready. This can be useful if you want
            to handle package additions identically whether they were already in the ledger when
            the client started up or only after a package has been added. The default value is
            ``False``, which means that this handler is only called on NEW packages that have been
            uploaded after this client has started.
        :return:
        """

        @wraps(handler)
        def _background_ledger_packages_added(
            event: PackagesAddedEvent,
        ) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        for key in EventKey.packages_added(initial=initial, changed=True):
            self._impl.add_event_handler(key, _background_ledger_packages_added, None, self)

    def ledger_transaction_start(self) -> EventHandlerDecorator[TransactionStartEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` receives a
        new transaction. Called before individual :meth:`ledger_create` and :meth:`ledger_archive`
        callbacks.
        """
        return fluentize(self.add_ledger_transaction_start)

    def add_ledger_transaction_start(self, handler: EventHandler[TransactionStartEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` receives a new transaction.
        Called before individual :meth:`ledger_create` and :meth:`ledger_archive` callbacks.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        """

        @wraps(handler)
        def _background_ledger_transaction_start(
            event: TransactionStartEvent,
        ) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        for key in EventKey.transaction_start():
            self._impl.add_event_handler(key, _background_ledger_transaction_start, None, self)

    def ledger_transaction_end(self) -> EventHandlerDecorator[TransactionEndEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` receives a
        new transaction. Called after individual :meth:`ledger_create` and :meth:`ledger_archive`
        callbacks.
        """

        def _register_transaction_end(
            cb: EventHandler[TransactionEndEvent],
        ) -> EventHandler[TransactionEndEvent]:
            self.add_ledger_transaction_end(cb)
            return cb

        return _register_transaction_end

    def add_ledger_transaction_end(self, handler: EventHandler[TransactionEndEvent]) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` receives a new transaction.
        Called after individual :meth:`ledger_create` and :meth:`ledger_archive` callbacks.

        :param handler:
            The handler to register. This can either be a coroutine or a normal function, and may
            return anything that can be successfully coerced into a :class:`CommandPayload`.
        """

        @wraps(handler)
        def _background_ledger_transaction_end(
            event: TransactionEndEvent,
        ) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        for key in EventKey.transaction_end():
            self._impl.add_event_handler(key, _background_ledger_transaction_end, None, self)

    def ledger_created(
        self, template: Any, match: Optional[ContractMatch] = None
    ) -> EventHandlerDecorator[ContractCreateEvent]:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters a newly created
        template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """

        def _register_created(
            cb: EventHandler[ContractCreateEvent],
        ) -> EventHandler[ContractCreateEvent]:
            self.add_ledger_created(template, match=match, handler=cb)
            return cb

        return _register_created

    def add_ledger_created(
        self,
        template: Any,
        handler: EventHandler[ContractCreateEvent],
        match: Optional[ContractMatch] = None,
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters a newly created
        contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param handler:
            The callback to invoke whenever a matching template is created.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """

        @wraps(handler)
        def _background_ledger_contract_create(
            event: ContractCreateEvent,
        ) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        filter_fn = partial(is_match, match) if match is not None else None

        for key in EventKey.contract_created(True, template):
            self._impl.add_event_handler(key, _background_ledger_contract_create, filter_fn, self)

    def ledger_exercised(
        self, template: "Any", choice: str
    ) -> EventHandlerDecorator[ContractExercisedEvent]:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters an exercised
        choice event.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param choice:
            The name of the choice to listen for exercises on.
        """

        def _register_exercised(
            cb: EventHandler[ContractExercisedEvent],
        ) -> EventHandler[ContractExercisedEvent]:
            self.add_ledger_exercised(template, choice, handler=cb)
            return cb

        return _register_exercised

    def add_ledger_exercised(
        self, template: Any, choice: str, handler: EventHandler[ContractExercisedEvent]
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters an exercised
        choice event.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param choice:
            The name of the choice to listen for exercises on.
        :param handler:
            The callback to invoke whenever a matching template is exercised.
        """

        @wraps(handler)
        def _background_ledger_contract_exercised(
            event: "ContractExercisedEvent",
        ) -> "Awaitable[EventHandlerResponse]":
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        for key in EventKey.contract_exercised(True, template, choice):
            self._impl.add_event_handler(key, _background_ledger_contract_exercised, None, self)

    def ledger_archived(
        self, template: Any, match: Optional[ContractMatch] = None
    ) -> EventHandlerDecorator[ContractArchiveEvent]:
        """
        Decorator for registering a callback to be invoked when the :class:`PartyClient` encounters
        a newly archived contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """

        def _register_archived(
            cb: EventHandler[ContractArchiveEvent],
        ) -> EventHandler[ContractArchiveEvent]:
            self.add_ledger_archived(template, match=match, handler=cb)
            return cb

        return _register_archived

    def add_ledger_archived(
        self,
        template: Any,
        handler: EventHandler[ContractArchiveEvent],
        match: Optional[ContractMatch] = None,
    ) -> None:
        """
        Register a callback to be invoked when the :class:`PartyClient` encounters a newly archived
        contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param handler:
            The callback to invoke whenever a matching template is created.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """

        @wraps(handler)
        def _background_ledger_contract_archived(
            event: ContractArchiveEvent,
        ) -> Awaitable[EventHandlerResponse]:
            return self._impl.invoker.run_in_executor(lambda: handler(event))

        filter_fn = partial(is_match, match) if match is not None else None

        for key in EventKey.contract_archived(True, template):
            self._impl.add_event_handler(key, _background_ledger_contract_archived, filter_fn, self)

    # </editor-fold>

    # region Command submission

    def submit(
        self,
        commands,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> None:
        """
        Submit commands to the ledger.

        :param commands:
            An object that can be converted to a command.
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        :return:
            A future that resolves when the command has made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        return self._impl.invoker.run_in_loop(
            lambda: self._impl.write_commands(
                commands, workflow_id=workflow_id, deduplication_time=deduplication_time
            )
        )

    def submit_create(
        self,
        template_name: TemplateNameLike,
        arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> None:
        """
        Synchronously submit a single create command. Equivalent to calling :meth:`submit` with a
        single ``create``.

        :param template_name:
            The name of the template.
        :param arguments:
            The arguments to the create (as a ``dict``).
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        """
        from .. import create

        return self.submit(
            create(template_name, arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    def submit_exercise(
        self,
        cid: ContractId,
        choice_name: str,
        arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> None:
        """
        Synchronously submit a single exercise choice. Equivalent to calling :meth:`submit` with a
        single ``exercise``.

        :param cid:
            The :class:`ContractId` on which a choice is being exercised.
        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            The arguments to the exercise (as a ``dict``). Can be omitted (``None``) for no-argument
            choices.
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        """
        from .. import exercise

        return self.submit(
            exercise(cid, choice_name, arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    def submit_exercise_by_key(
        self,
        template_name: TemplateNameLike,
        contract_key: Any,
        choice_name: str,
        arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> None:
        """
        Synchronously submit a single exercise choice. Equivalent to calling :meth:`submit` with a
        single ``exercise_by_key``.

        :param template_name:
            The name of the template on which to do an exercise-by-key.
        :param contract_key:
            The value that should uniquely identify a contract for the specified template.
        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            The arguments to the create (as a ``dict``). Can be omitted (``None``) for no-argument
            choices.
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        """
        from .. import exercise_by_key

        return self.submit(
            exercise_by_key(template_name, contract_key, choice_name, arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    def submit_create_and_exercise(
        self,
        template_name: TemplateNameLike,
        arguments: dict,
        choice_name: str,
        choice_arguments: Optional[dict] = None,
        workflow_id: Optional[str] = None,
        deduplication_time: Optional[TimeDeltaLike] = None,
    ) -> None:
        """
        Synchronously submit a single create-and-exercise command. Equivalent to calling
        :meth:`submit` with a single ``create_and_exercise``.

        :param template_name:
            The name of the template on which to do an exercise-by-key.
        :param arguments:
            The arguments to the create (as a ``dict``).
        :param choice_name:
            The name of the choice to exercise.
        :param choice_arguments:
            The arguments to the exercise (as a ``dict``). Can be omitted (``None``) for no-argument
        :param workflow_id:
            The optional workflow ID to stamp on the outgoing command.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        """
        from .. import create_and_exercise

        return self.submit(
            create_and_exercise(template_name, arguments, choice_name, choice_arguments),
            workflow_id=workflow_id,
            deduplication_time=deduplication_time,
        )

    # endregion

    # <editor-fold desc="Active contract set">

    def find_by_id(self, cid: Union[str, ContractId]) -> Optional[ContractContextualData]:
        return self._impl.invoker.run_in_loop(lambda: self._impl.find_by_id(cid))

    def find(
        self, template: Any, match: ContractMatch = None, include_archived: bool = False
    ) -> ContractContextualDataCollection:
        return self._impl.invoker.run_in_loop(
            lambda: self._impl.find(template, match, include_archived=include_archived)
        )

    def find_active(self, template: Any, match: ContractMatch = None) -> "ContractsState":
        """
        Immediately return data from the current active contract set.

        The contents of this ACS are guaranteed to be present (or removed) in the current
        transaction _before_ processing any corresponding ``on_created`` or ``on_archived``
        callbacks for this party. The ACS is populated _before_ processing any ``on_ready``
        callbacks.

        This method raises an error if ACS tracking has been disabled on this client.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :return:
            A ``dict`` whose keys are :class:`ContractId` and values are corresponding contract
            data that match the current query.
        :return:
            A ``dict`` whose keys are :class:`ContractId` and values are corresponding contract
            data that match the current query.
        """
        return self._impl.invoker.run_in_loop(lambda: self._impl.find_active(template, match))

    def find_historical(
        self, template: Any, match: ContractMatch = None
    ) -> ContractContextualDataCollection:
        """
        Immediately return data from the current active and historical contract set.

        The contents of this set are guaranteed to be up-to-date in the current transaction _before_
        processing any corresponding ``on_created`` or ``on_archived`` callbacks for this party. The
        set is up-to-date _before_ processing any ``on_ready`` callbacks.

        This method raises an error if historical tracking has been disabled on this client.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :return:
            A ``dict`` whose keys are :class:`ContractId` and values are corresponding contract
            data that match the current query.
        :return:
            A ``dict`` whose keys are :class:`ContractId` and values are corresponding contract
            data that match the current query.
        """
        return self._impl.invoker.run_in_loop(lambda: self._impl.find_historical(template, match))

    def find_one(
        self, template: Any, match: ContractMatch = None, timeout: float = DEFAULT_TIMEOUT_SECONDS
    ) -> Tuple[ContractId, ContractData]:
        """
        Return data from the current active contract set when at least some amount of rows exist in
        the active contract set.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :param timeout:
            Number of seconds in which to time out the search.
        :return:
            A ``Future`` that is resolved with a ``dict`` whose keys are :class:`ContractId` and
            values are corresponding contract data that match the current query.
        """
        state = self.find_nonempty(template, match, min_count=1, timeout=timeout)
        return next(iter(state.items()))

    def find_nonempty(
        self,
        template: Any,
        match: ContractMatch,
        min_count: int = 1,
        timeout: float = DEFAULT_TIMEOUT_SECONDS,
    ) -> ContractsState:
        """
        Return data from the current active contract set when at least some amount of rows exist in
        the active contract set.

        :param template:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :param min_count:
            The minimum number of rows to return. The default value is 1.
        :param timeout:
            Number of seconds in which to time out the search.
        :return:
            A ``Future`` that is resolved with a ``dict`` whose keys are :class:`ContractId` and
            values are corresponding contract data that match the current query.
        """
        return self._impl.invoker.run_in_loop(
            lambda: self._impl.find_nonempty(template, match, min_count=min_count, timeout=timeout)
        )

    # </editor-fold>

    # <editor-fold desc="Ledger/client metadata">

    def set_config(self, url: Optional[str], **kwargs):
        self._impl.set_config(url=url, **kwargs)

    def ensure_dar(
        self,
        contents: Union[str, Path, bytes, BinaryIO],
        timeout: TimeDeltaLike = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        Validate that the ledger has the packages specified by the given contents (as a byte array).
        Throw an exception if the specified DARs do not exist within the specified timeout.

        :param contents: The DAR or DALF to ensure.
        :param timeout: The maximum length of time to wait before giving up.
        """
        raw_bytes = get_bytes(contents)
        return self._impl.invoker.run_in_loop(
            lambda: self._impl.parent.upload_package(raw_bytes, timeout)
        )

    def ready(self) -> None:
        """
        Block until the underlying infrastructure has connected to all necessary services.
        """
        # TODO: Improve on this implementation; this spin loop is unnecessarily ugly
        from time import sleep

        while self._impl.invoker.loop is None:
            sleep(0.1)

        LOG.debug("Waiting for the underlying implementation to be ready...")
        return self._impl.invoker.run_in_loop(lambda: self._impl.ready())

    # </editor-fold>



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from ..prim import DazlError

__all__ = [
    "ConnectionTimeoutError",
    "UserTerminateRequest",
    "StreamError",
    "ProtocolWarning",
    "CallbackReturnWarning",
]


class ConnectionTimeoutError(DazlError):
    """
    Raised when a connection failed to be established before the connection timeout elapsed.
    """


class UserTerminateRequest(DazlError):
    """
    Raised when the user has initiated a request to terminate the application.
    """


class StreamError:
    """
    An error that arises when trying to read from a query stream.
    """


class ProtocolWarning(Warning):
    """
    Warnings that are raised when dazl detects incompatibilities between the Ledger API server-side
    implementation and dazl.
    """



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Sequence


def unmangle_name(s: str) -> str:
    """
    Convert a name with ``$uXXXX`` character runs into their original forms, which may not
    necessarily be safe DAML identifiers.

    :param s: The string to interpret.
    :return: A new string with the appropriate substitutions.
    """
    return (
        s.replace("$u002b", "+")
        .replace("$u005b", "[")
        .replace("$u005d", "]")
        .replace("$u003c", "<")
        .replace("$u003e", ">")
        .replace("$u003a", ":")
        .replace("$u0022", '"')
        .replace("$u0028", "(")
        .replace("$u0029", ")")
        .replace("$u002f", "/")
        .replace("$u002c", ",")
        .replace("$u003d", "=")
    )


def is_hidden_module_name(name: "Sequence[str]") -> bool:
    if len(name) == 1 and name[0] in ("GhcPrim", "Prelude"):
        return True
    if len(name) >= 2:
        if name[0] == "GHC":
            return True
        elif name[0] == "DA":
            return True
        elif name[0] == "Control" and name[1] == "Exception":
            return True
    return False


def maybe_parentheses(obj, operator: str = " ") -> str:
    """
    Wrap the string (or object's string representation) in parentheses, but only if required.
    :param obj:
        The object to (potentially) wrap up in parentheses. If it is not a string, it is converted
        to a string first.
    :param operator:
        The operator that needs precedence rules clarified. The default is ``' '`` (a space).
    :return: The (possibly) parenthesized string.
    """
    s = str(obj)
    if operator not in s:
        return s

    # if there are already wrapping punctuation marks, there may not be a need to add additional
    # ones
    open_mark = s[0]
    if open_mark in "([{":
        groupings = [s[0]]
        for c in s[1:-1]:
            if c in "([{":
                groupings.append(c)
            elif c == "}":
                if groupings[-1] == "{":
                    groupings.pop()
                else:
                    groupings.clear()
            elif c == "]":
                if groupings[-1] == "[":
                    groupings.pop()
                else:
                    groupings.clear()
            elif c == ")":
                if groupings[-1] == "(":
                    groupings.pop()
                else:
                    groupings.clear()

            if not groupings:
                # we balanced out all groupings (or just his a syntax error in unmatched groupings);
                # add clarifying parentheses
                return "(" + s + ")"
        if (
            (groupings[0] == "(" and groupings[-1] == ")")
            or (groupings[0] == "[" and groupings[-1] == "]")
            or (groupings[0] == "{" and groupings[-1] == "}")
        ):
            return s
    return "(" + s + ")" if operator in s else s


def indent(text: str, spaces: int):
    """
    Prepend every line of the specified text with a set number of spaces. Line endings are
    preserved.
    """
    prefix = " " * spaces
    return "".join(prefix + t for t in text.splitlines(keepends=True))



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import ensure_future, gather, get_event_loop, sleep, wait_for
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta
import sys
from typing import AbstractSet, Awaitable, Callable, Dict, Optional, Set, TypeVar

from .. import LOG
from ..damlast.daml_lf_1 import Archive, Package, PackageRef
from ..damlast.errors import NameNotFoundError, PackageNotFoundError
from ..damlast.lookup import MultiPackageLookup, validate_template
from ..damlast.pkgfile import Dar
from ..prim import DazlError

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

__all__ = ["PackageService", "PackageLoader", "DEFAULT_TIMEOUT"]

T = TypeVar("T")

# mypy insists on having a type annotation here, or it will complain about not being able to
# determine the type of this field in pkgloader_aio_compat.py
DEFAULT_TIMEOUT: timedelta = timedelta(seconds=30)


class PackageService(Protocol):
    """
    A service that provides package information.
    """

    async def get_package(self, package_id: "PackageRef") -> bytes:
        raise NotImplementedError("SyncPackageService.package_bytes requires an implementation")

    async def list_package_ids(self) -> "AbstractSet[PackageRef]":
        raise NotImplementedError("SyncPackageService.package_ids requires an implementation")


class PackageLoader:
    """
    Loader for packages from a remote PackageService.

    This class handles retries and backoffs, and avoids having more than one request in flight for
    the same package ID. It is intended to be shared by all local clients that may need package
    information.
    """

    _allow_deprecated_identifiers = False

    def __init__(
        self,
        package_lookup: "MultiPackageLookup",
        conn: "Optional[PackageService]" = None,
        timeout: "Optional[timedelta]" = DEFAULT_TIMEOUT,
        executor: "Optional[ThreadPoolExecutor]" = None,
    ):
        self._package_lookup = package_lookup
        self._conn = conn
        self._timeout = timeout or DEFAULT_TIMEOUT
        self._loading_futs = dict()  # type: Dict[PackageRef, Awaitable[Package]]
        self._parsing_futs = dict()  # type: Dict[PackageRef, Awaitable[Archive]]
        self._executor = executor or ThreadPoolExecutor(3)

    def set_connection(self, conn: "Optional[PackageService]"):
        self._conn = conn

    async def do_with_retry(self, fn: "Callable[[], T]") -> "T":
        """
        Perform a synchronous action that assumes the existence of one or more packages. In the
        event the function raises :class:`PackageNotFoundError` or a wildcarded
        :class:`NameNotFoundError`, the required package/type is fetched and the operation retried.

        If, after a retry, an expected package or type could not be found, the exception is
        re-raised to the caller.

        :param fn: A function to invoke.
        :return: The result of that function.
        """
        failed_types = set()  # type: Set[str]
        failed_packages = set()  # type: Set[PackageRef]
        while True:
            try:
                return fn()

            except PackageNotFoundError as ex:
                # every time we fail serialization due to a missing package or type,
                # try to resolve it; remember what we tried, because if we fail again
                # for the same reason it is likely fatal
                if ex.ref in failed_packages:
                    # we already looked for this package and couldn't find it; this will
                    # never succeed
                    raise
                failed_packages.add(ex.ref)
                await self.load(ex.ref)

            except NameNotFoundError as ex:
                if ex.ref in failed_types:
                    # we already looked for this type and couldn't find it; this will
                    # never succeed
                    LOG.verbose(
                        "Failed to find name %s in all known packages, "
                        "even after fetching the latest.",
                        ex.ref,
                    )
                    raise

                pkg_id, name = validate_template(
                    ex.ref, allow_deprecated_identifiers=self._allow_deprecated_identifiers
                )
                if pkg_id == "*":
                    # we don't know what package contains this type, so we have no
                    # choice but to look in all known packages
                    LOG.verbose(
                        "Failed to find name %s in all known packages, "
                        "so loading ALL packages...",
                        name,
                    )
                    failed_types.add(ex.ref)
                    await self.load_all()
                else:
                    # we know what package this type comes from, but it did not contain
                    # the required type
                    LOG.warning("Found package %s, but it did not include type %s", pkg_id, name)
                    raise

    async def preload(self, *contents: "Dar") -> None:
        """
        Populate a :class:`PackageCache` with types from DARs.

        :param contents:
            One or more DARs to load into a local package cache.
        """

    async def load(self, ref: "PackageRef") -> "Package":
        """
        Load a package ID from the remote server. If the package has additional dependencies, they
        are also loaded.

        :param ref: One or more :class:`PackageRef`s.
        :raises: PackageNotFoundError if the package could not be resolved
        """
        # If the package has already been loaded, then skip all the expensive I/O stuff
        try:
            return self._package_lookup.package(ref)
        except PackageNotFoundError:
            pass

        # If we already have a request in-flight, simply return that same Future to our caller;
        # do not try to schedule a new request
        fut = self._loading_futs.get(ref)
        if fut is None:
            fut = ensure_future(self._load_and_parse_package(ref))
            self._loading_futs[ref] = fut
        package = await fut

        _ = self._loading_futs.pop(ref, None)
        _ = self._parsing_futs.pop(ref, None)

        return package

    async def _load_and_parse_package(self, package_id: "PackageRef") -> "Package":
        from ..damlast.parse import parse_archive

        LOG.info("Loading package: %s", package_id)

        loop = get_event_loop()
        conn = self._conn
        if conn is None:
            raise DazlError("a connection is not configured")

        archive_bytes = await wait_for(
            self.__fetch_package_bytes(conn, package_id), timeout=self._timeout.total_seconds()
        )

        LOG.info("Loaded for package: %s, %d bytes", package_id, len(archive_bytes))

        # we only ever want a package to be parsed once; it could be that there were multiple
        # attempts to load a package in flight (though this shouldn't happen either)
        fut = self._parsing_futs.get(package_id)
        if fut is None:
            fut = ensure_future(
                loop.run_in_executor(
                    self._executor, lambda: parse_archive(package_id, archive_bytes)
                )
            )
            self._parsing_futs[package_id] = fut

        archive = await fut
        self._package_lookup.add_archive(archive)
        return archive.package

    @staticmethod
    async def __fetch_package_bytes(conn: "PackageService", package_id: "PackageRef") -> bytes:
        sleep_interval = 1

        while True:
            # noinspection PyBroadException
            try:
                return await conn.get_package(package_id)
            except Exception:
                # We tried fetching the package but got an error. Retry, backing off to waiting as
                # much as 30 seconds between each attempt.
                await sleep(sleep_interval)
                sleep_interval = min(sleep_interval * 2, 30)
                LOG.exception("Failed to fetch package; this will be retried.")

    async def load_all(self):
        """
        Load all packages from the remote server.
        """
        package_ids = set(await self._conn.list_package_ids())
        package_ids -= self._package_lookup.package_ids()
        if package_ids:
            await gather(*(self.load(package_id) for package_id in package_ids))



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# type: ignore

from typing import Mapping, Optional, Sequence, Union

from ..damlast.daml_lf_1 import (
    Block,
    BuiltinFunction,
    Case,
    DefDataType,
    Expr,
    Module,
    PrimCon,
    PrimLit,
    PrimType,
    Scenario,
    Type,
    Update,
    ValName,
)
from ..damlast.util import pack_arrow_type, unpack_arrow_type
from ._render_base import (
    CodeContext,
    PrettyPrintBase,
    decode_special_chars,
    register_pretty_printer,
)


class CSharpPrettyPrint(PrettyPrintBase):

    syntax = "csharp"

    def lexer(self):
        try:
            from pygments.lexers.dotnet import CSharpLexer
        except ImportError:
            return None
        return CSharpLexer()

    def visit_module(self, module: "Module") -> str:
        prefix = (
            f'namespace {".".join(module.name.segments[:-1])} {{\n'
            + f"public static class {module.name.segments[-1]} {{\n"
        )

        lines = []
        for value in module.values:
            proposed_name = ".".join(value.name_with_type.name)
            final_name = decode_special_chars(proposed_name)
            if final_name != proposed_name:
                final_name = f"{proposed_name} ({final_name})"

            lines.append("/*")
            lines.append(" NAME: " + final_name)
            lines.append(" TYPE: " + repr(value.name_with_type.type))
            lines.append(" EXPR: " + repr(value.expr))
            lines.append("*/")
            child = self.with_decl(value.name_with_type.name, value.name_with_type.type)
            expr_text = child.visit_expr(value.expr)
            if expr_text is None:
                type_text = self.visit_type(value.name_with_type.type)
                expr_text = f'// {value.expr!r}\npublic extern {type_text} {".".join(value.name_with_type.name)};'
            lines.extend(expr_text.splitlines())
            lines.append("")

        for data_type in module.data_types:
            def_data_text = self.visit_def_data_type(data_type)
            lines.extend(def_data_text.splitlines())
            lines.append("")

        return (
            prefix
            + "\n".join("    " + line.replace("$", "_") if line else "" for line in lines)
            + "}}\n\n"
        )

    def visit_def_data_type(self, def_data_type: "DefDataType") -> "str":
        lines = []
        type_params = [v.var for v in def_data_type.params]
        type_param_text = ("<" + ", ".join(type_params) + ">") if type_params else ""
        if def_data_type.serializable:
            lines.append("[Serializable]")
        if def_data_type.record is not None:
            fields = {f.field: self.visit_type(f.type) for f in def_data_type.record.fields}
            lines.append(
                "public sealed class "
                + ".".join(def_data_type.name.segments)
                + type_param_text
                + " {"
            )
            for name, type_str in fields.items():
                lines.append(f"    public {type_str} {name} {{ get; }} = {name};")
            lines.append("}")
        elif def_data_type.variant is not None:
            fields = {f.field: self.visit_type(f.type) for f in def_data_type.variant.fields}
            lines.append(
                "public sealed class "
                + ".".join(def_data_type.name.segments)
                + type_param_text
                + " {"
            )
            for name, type_str in fields.items():
                lines.append(f"    public {type_str} {name} {{ get; }} = {name};")
            lines.append("}")
        else:
            lines.append("public sealed class " + ".".join(def_data_type.name.segments) + "{}")
        return "\n".join(lines) + "\n"

    def visit_daml_template(self, template: "DamlTemplate"):

        pass

    def visit_expr_abs_decl_eq(self, expr: "Expr.Abs"):
        return f"""
        /**
         * Originally defined at ...
         */
        public static boolean operator ==() {{
        }}"""

    def visit_expr_var(self, var: str) -> "str":
        return self._render_rhs(var, None)

    def visit_expr_val(self, val: "ValName") -> "str":
        if self.context.decl_name is not None:
            lines = []
            if self.context.type_abs is not None:
                lines.append(f"// WARNING: Not sure how to render these: {self.context.type_abs}")
            lines.append(
                f"public static {self.visit_type(self.context.decl_type)} "
                f'{".".join(self.context.decl_name)} = {val.full_name}'
            )
            return "\n".join(lines)
        else:
            return self._render_rhs(val.full_name, None)

    def visit_expr_builtin(self, builtin: "BuiltinFunction") -> "str":
        return self._render_rhs(f"BuiltIns.{builtin.name}", None)

    def visit_expr_prim_con(self, prim_con: "PrimCon") -> "str":
        if PrimCon.CON_TRUE == prim_con:
            text = "true"
        elif PrimCon.CON_FALSE == prim_con:
            text = "false"
        elif PrimCon.CON_UNIT == prim_con:
            text = "Unit"
        else:
            text = repr(prim_con)
        return self._render_rhs(text, None)

    def visit_expr_prim_lit(self, prim_lit: "PrimLit") -> "str":
        if prim_lit.int64 is not None:
            text = str(prim_lit.int64)
        elif prim_lit.party is not None:
            text = f'Party("{prim_lit.party}")'
        elif prim_lit.text is not None:
            import json

            text = json.dumps(prim_lit.text)
        elif prim_lit.date is not None:
            from datetime import date, timedelta

            d = date(1970, 1, 1) + timedelta(days=prim_lit.date)
            text = f"Date({d.year}, {d.month}, {d.day})"
        elif prim_lit.timestamp is not None:
            text = f"ConstructDateTime({prim_lit.timestamp})"
        else:
            text = repr(prim_lit)

        return self._render_rhs(text, None)

    def _render_rhs(self, ctor: str, args: Optional[Sequence[str]]):
        if self.context.decl_name is not None:
            lines = []
            if self.context.type_abs is not None:
                lines.append(f"// WARNING: Not sure how to render these: {self.context.type_abs}")

            decl_line = (
                f"public static {self.visit_type(self.context.decl_type)} "
                f'{".".join(self.context.decl_name)} = {ctor}'
            )
            if args is not None:
                decl_line += "("
                delim = ""
                for arg in args:
                    decl_line += delim
                    delim = ","
                    decl_line += "\n    "
                    decl_line += arg
                decl_line += ")"
            decl_line += ";"
            return decl_line
        else:
            decl_line = ctor
            if args is not None:
                delim = "("
                for arg in args:
                    decl_line += delim
                    delim = ", "
                    decl_line += arg
                decl_line += ")"

        return decl_line if self.context.in_expression else f"return {decl_line};"

    def visit_expr_rec_con(self, rec_con: "Expr.RecCon") -> "str":
        scope = self.with_expression()
        args = [f"{field.field}: {scope.visit_expr(field.expr)}" for field in rec_con.fields]
        return self._render_rhs("new " + self.visit_type_con(rec_con.tycon), args)

    def visit_expr_rec_proj(self, rec_proj: "Expr.RecProj") -> "str":
        scope = self.with_expression()
        return self._render_rhs(f"{scope.visit_expr(rec_proj.record)}.{rec_proj.field}", None)

    def visit_expr_variant_con(self, variant_con: "Expr.VariantCon") -> "str":
        scope = self.with_expression()
        args = [f"{variant_con.variant_con}: {scope.visit_expr(variant_con.variant_arg)}"]
        return self._render_rhs(self.visit_type_con(variant_con.tycon), args)

    def visit_expr_struct_con(self, struct_con: "Expr.StructCon") -> "str":
        scope = self.with_expression()
        args = [f"{field}: {scope.visit_expr(field.expr)}" for field in struct_con.fields]
        return self._render_rhs("", args)

    def visit_expr_struct_proj(self, struct_proj: "Expr.StructProj") -> "str":
        scope = self.with_expression()
        return self._render_rhs(f"{scope.visit_expr(struct_proj.tuple)}.{struct_proj.field}", None)

    def visit_expr_app(self, app: "Expr.App") -> "str":
        scope = self.with_expression()
        ctor_name = scope.visit_expr(app.fun)
        if self.context.decl_name is not None:
            lines = []
            if self.context.type_abs is not None:
                lines.append(f"// WARNING: Not sure how to render these: {self.context.type_abs}")

            if self.context.type_app:
                type_app_str = (
                    "<" + ", ".join(self.visit_type(t) for t in self.context.type_app) + ">"
                )
            else:
                type_app_str = ""
            decl_line = (
                f"public static {self.visit_type(self.context.decl_type)} "
                f'{".".join(self.context.decl_name)} = {ctor_name}{type_app_str}('
            )
            delim = ""
            for arg in app.args:
                decl_line += delim
                delim = ","
                decl_line += "\n    "
                decl_line += scope.visit_expr(arg)
            decl_line += ");"

            return decl_line
        else:
            decl_line = ctor_name
            if " " in decl_line:
                decl_line = "(" + decl_line + ")"
            delim = "("
            for arg in app.args:
                decl_line += delim
                delim = ", "
                decl_line += scope.visit_expr(arg)
            decl_line += ")"

            return decl_line if self.context.in_expression else f"return {decl_line};"

    def visit_expr_abs_decl(self, abs_: "Expr.Abs", context: "CodeContext") -> "str":
        fn_types = unpack_arrow_type(context.decl_type)

        if context.type_abs is not None:
            generic_args = (
                "<" + (", ".join(self.visit_type_var(tvwk.var) for tvwk in context.type_abs)) + ">"
            )
        else:
            generic_args = ""

        length_of_abstraction = len(abs_.param)
        length_of_proposed_signature = len(fn_types) - 1

        fn_args = {}  # type: Mapping[str, Type]
        ret_type = None  # type: Type
        if length_of_abstraction <= length_of_proposed_signature:
            # TODO: vwt.type and type should always agree with each other; might be nice if we validated that
            #  here instead of just assuming
            fn_args = {vwt.var: self.visit_type(type) for vwt, type in zip(abs_.param, fn_types)}
            ret_type = pack_arrow_type(fn_types[len(abs_.param) :])
        elif length_of_abstraction > length_of_proposed_signature:
            # we're passing additional parameters to function returned from the original expression
            fn_args = {vwt.var: self.visit_type(type) for vwt, type in zip(abs_.param, fn_types)}
            return "???"

        fn_args_text = ", ".join(
            f"{arg_type_text} {arg_name}" for arg_name, arg_type_text in fn_args.items()
        )

        child = self.with_statement_block()
        return_type_text = self.visit_type(ret_type)
        l1 = f'public static {return_type_text} {".".join(context.decl_name)}{generic_args}({fn_args_text}) {{'
        l2 = child.visit_expr(abs_.body) or "..."
        l3 = "}"
        return "\n".join([l1, *("    " + text for text in l2.splitlines()), l3])

    def visit_expr_abs_inline(self, abs_: "Expr.Abs") -> "str":
        type_params = []
        for param in abs_.param:
            type_expr = self.visit_type(param.type) + " "
            if type_expr == "var ":
                type_expr = ""
            type_params.append(type_expr + param.var)
        body_expr = self.visit_expr(abs_.body)
        expr_text = f'({", ".join(type_params)}) => {body_expr}'
        return expr_text if self.context.in_expression else f"return {expr_text};"

    def visit_expr_ty_abs(self, ty_abs: "Expr.TyAbs") -> "str":
        return self.with_type_abs(ty_abs.param).visit_expr(ty_abs.body)

    def visit_expr_case(self, case: "Case") -> "str":
        pass

    def visit_expr_let(self, let: "Block") -> "str":
        lines = []
        scope = self.with_expression()
        for binding in let.bindings:
            type_str = scope.visit_type(binding.binder.type)
            expr_str = scope.visit_expr(binding.bound)
            lines.append(f"{type_str} {binding.binder.var} = {expr_str};")
        lines.append("return " + scope.visit_expr(let.body) + ";")
        return "\n".join(lines)

    def visit_expr_nil(self, nil: "Expr.Nil") -> "str":
        return self._render_rhs(f"ImmutableList<{self.visit_type(nil.type)}>.Empty", None)

    def visit_expr_cons(self, cons: "Expr.Cons") -> "str":
        scope = self.with_expression()
        front_args = ", ".join(scope.visit_expr(f) for f in cons.front)
        tail_arg = scope.visit_expr(cons.tail)
        expr_text = f"ImmutableList.Create<{scope.visit_type(cons.type)}>({front_args}).AddRange({tail_arg})"
        return self._render_rhs(expr_text, None)

    def visit_expr_update(self, update: "Update") -> "str":
        pass

    def visit_expr_scenario(self, scenario: "Scenario") -> "str":
        pass

    def visit_expr_rec_upd(self, rec_upd: "Expr.RecUpd") -> "str":
        pass

    def visit_expr_struct_upd(self, struct_upd: "Expr.StructUpd") -> "str":
        pass

    def visit_expr_optional_none(self, optional_none: "Expr.OptionalNone") -> "str":
        expr_text = f"ImmutableList<{self.visit_type(optional_none.type)}>.Empty"
        return expr_text if self.context.in_expression else f"return {expr_text};"

    def visit_expr_optional_some(self, optional_some: "Expr.OptionalSome") -> "str":
        return self.visit_expr(optional_some.body)

    def visit_type_var(self, var: "Union[str, Type.Var]") -> "str":
        rendered_var = super(CSharpPrettyPrint, self).visit_type_var(var)
        return rendered_var[0].upper() + rendered_var[1:]

    def visit_type_con(self, con: "Type.Con") -> "str":
        type_args = [self.visit_type(arg) for arg in con.args]
        if type_args:
            return f'{con.tycon.full_name}<{", ".join(type_args)}>'
        else:
            return con.tycon.full_name

    def visit_type_prim(self, prim: "Type.Prim") -> "str":
        if PrimType.UNIT == prim.prim:
            return "Unit"
        elif PrimType.BOOL == prim.prim:
            return "boolean"
        elif PrimType.INT64 == prim.prim:
            return "int"
        elif PrimType.DECIMAL == prim.prim:
            return "Decimal"
        elif PrimType.CHAR == prim.prim:
            return "string"
        elif PrimType.TEXT == prim.prim:
            return "string"
        elif PrimType.TIMESTAMP == prim.prim:
            return "DateTime"
        elif PrimType.RELTIME == prim.prim:
            return "timedelta"
        elif PrimType.PARTY == prim.prim:
            return "Party"
        elif PrimType.LIST == prim.prim:
            if len(prim.args) == 0:
                return "List<>"
            else:
                return f"List<{self.visit_type(prim.args[0])}>"
        elif PrimType.UPDATE == prim.prim:
            if len(prim.args) == 0:
                return "Update<>"
            else:
                return f"Update<{self.visit_type(prim.args[0])}>"
        elif PrimType.SCENARIO == prim.prim:
            if len(prim.args) == 0:
                return "Scenario"
            else:
                return f"Scenario<{self.visit_type(prim.args[0])}>"
        elif PrimType.DATE == prim.prim:
            return "Date"
        elif PrimType.CONTRACT_ID == prim.prim:
            return f"ContractId<{self.visit_type(prim.args[0])}>"
        elif PrimType.OPTIONAL == prim.prim:
            if len(prim.args) == 0:
                return "Optional<>"
            else:
                return f"Optional<{self.visit_type(prim.args[0])}>"
        elif PrimType.ARROW == prim.prim:
            type_strings = [self.visit_type(a) for a in prim.args]
            return f'Func<{", ".join(type_strings)}>'
        elif PrimType.TEXTMAP == prim.prim:
            return f"Dictionary<{self.visit_type(prim.args[0])},  {self.visit_type(prim.args[1])}>"
        else:
            raise ValueError(f"unknown Type.Prim: {prim!r}")

    def visit_type_forall(self, forall: "Type.Forall") -> "str":
        return self.visit_type(forall.body)

    def visit_type_tuple(self, tuple: "Type.Tuple") -> "str":
        args = ", ".join(f"{self.visit_type(fwt.type)} {fwt.field}" for fwt in tuple.fields)
        return f"({args})"

    def _visit_type_default(self):
        return "var"



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    Mapping as __Mapping,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper as __EnumTypeWrapper
from google.protobuf.message import Message as __Message

from .......damlast.daml_lf_1 import PackageRef
from .trace_context_pb2 import TraceContext

__all__ = [
    "ListPackagesRequest",
    "ListPackagesResponse",
    "GetPackageRequest",
    "GetPackageResponse",
    "GetPackageStatusRequest",
    "GetPackageStatusResponse",
    "HashFunction",
    "SHA256",
    "PackageStatus",
    "UNKNOWN",
    "REGISTERED",
]

class ListPackagesRequest(__Message):
    ledger_id: str
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ListPackagesRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ListPackagesResponse(__Message):
    package_ids: __List[PackageRef]
    def __init__(self, *, package_ids: __Optional[__Sequence[PackageRef]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ListPackagesResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["package_ids"]) -> bool: ...
    def ClearField(self, field_name: __Literal["package_ids"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetPackageRequest(__Message):
    ledger_id: str
    package_id: PackageRef
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., package_id: __Optional[PackageRef] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetPackageRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "package_id", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "package_id", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetPackageResponse(__Message):
    hash_function: __Literal[0]
    archive_payload: bytes
    hash: PackageRef
    def __init__(self, *, hash_function: __Optional[__Literal[0]] = ..., archive_payload: __Optional[bytes] = ..., hash: __Optional[PackageRef] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetPackageResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["hash_function", "archive_payload", "hash"]) -> bool: ...
    def ClearField(self, field_name: __Literal["hash_function", "archive_payload", "hash"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetPackageStatusRequest(__Message):
    ledger_id: str
    package_id: PackageRef
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., package_id: __Optional[PackageRef] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetPackageStatusRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "package_id", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "package_id", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetPackageStatusResponse(__Message):
    package_status: __Literal[0, 1]
    def __init__(self, *, package_status: __Optional[__Literal[0, 1]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetPackageStatusResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["package_status"]) -> bool: ...
    def ClearField(self, field_name: __Literal["package_status"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class HashFunction(__EnumTypeWrapper):
    @overload
    def Name(self, number: __Literal[0]) -> __Literal["SHA256"]: ...
    @overload
    def Value(self, name: __Literal["SHA256"]) -> __Literal[0]: ...

SHA256: __Literal[0] = ...

class PackageStatus(__EnumTypeWrapper):
    @overload
    def Name(self, number: __Literal[0]) -> __Literal["UNKNOWN"]: ...
    @overload
    def Name(self, number: __Literal[1]) -> __Literal["REGISTERED"]: ...
    @overload
    def Value(self, name: __Literal["UNKNOWN"]) -> __Literal[0]: ...
    @overload
    def Value(self, name: __Literal["REGISTERED"]) -> __Literal[1]: ...

UNKNOWN: __Literal[0] = ...
REGISTERED: __Literal[1] = ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any as __Any,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    Mapping as __Mapping,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.message import Message as __Message

from .......prim import Party
from .value_pb2 import Identifier

__all__ = ["TransactionFilter", "Filters", "InclusiveFilters"]

class TransactionFilter(__Message):
    filters_by_party: __Map[Party, Filters]
    def __init__(self, *, filters_by_party: __Optional[__Mapping[Party, Filters]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: TransactionFilter) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["filters_by_party"]) -> bool: ...
    def ClearField(self, field_name: __Literal["filters_by_party"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Filters(__Message):
    inclusive: InclusiveFilters
    def __init__(self, *, inclusive: __Optional[InclusiveFilters] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Filters) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["inclusive"]) -> bool: ...
    def ClearField(self, field_name: __Literal["inclusive"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class InclusiveFilters(__Message):
    template_ids: __List[Identifier]
    def __init__(self, *, template_ids: __Optional[__Sequence[Identifier]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: InclusiveFilters) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["template_ids"]) -> bool: ...
    def ClearField(self, field_name: __Literal["template_ids"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...
